;
; Routines for realtime delays without using a timer
;
; Andreas Tnne
; atoenne@t-online.de
;
; No copyright. This is free software. Thank you for naming the author
;
; Web released version 1.0
;
;***********************************************************************

;*********************   USAGE   ***************************************
;
; .DEF DelayInnerLoop	Parameters for ShortDelay
; .DEF DelayOuterLoop	see the comment for ShortDelay
;
; rcall ShortDelay		wait for (DelayInnerLoop*5 + 1) * DelayOuterLoop + 13 clocks
;						In: DelayInnerLoop, DelayOuterLoop
; Delay Time			Same as ShortDelay but the time is given in microseconds
;						Modifies: DelayInnerLoop, DelayOuterLoop
;
;***********************************************************************


;#######################################################################

; ****** ShortDelay ******
; run a short delay using InnerLoop and OuterLoop
;
; timing calculation
; clocks = (InnerLoop * 5 + 10) * OuterLoop + 5
; Example: 	at 4MHz each clock is 1/4 us.
;			each InnerLoop is then 1.25us plus 2.5us overhead
; To achieve 250us one can use
; 	InnerLoop 18 (inner loop will be 25us = 100clocks) and
;	OuterLoop 10 (ignoring the extra 4us for overhead and calling 
; To achive 15ms one can use
;	InnerLoop 118 = 150us
;	OuterLoop 100 (ignoring the extra 4us for overhead)
; The maximum is approx. 82ms at 4Mhz

ShortDelay:

RunOuterLoop:					; (5*IL+1 + 9) * OuterLoop + 1
	push DelayInnerLoop			; #2 save it for next run of outer loop

RunInnerLoop:					; 5 * InnerLoop + 1
	dec DelayInnerLoop			; #1
	cpi DelayInnerLoop, 0		; #1
	nop							; #1
	nop							; #1
	brne RunInnerLoop			; #1/2

	pop DelayInnerLoop			; #2 prepare next run, clean stack
	dec DelayOuterLoop			; #1
	cpi DelayOuterLoop, 0		; #1
	nop							; #1
	nop							; #1
	brne RunOuterLoop			; #1/2

	ret							; #4

; ****** Delay Time ******
; Macro that waits aprox Time microseconds
; Note: we use the scaled clock value to avoid a 32bit overflow in the macro expression
; The computation is as follow:
; The NUMBER of iterations through the inner loop is precisely TIME * CLOCK / (ClockPerInnerLoop * 1000000)
; The number of OUTERLOOP runs is NUMBER / (255 + 2) (+2 for overhead which is equivalent to two inner loop runs)
; The number of INNTERLOOP runs NUMBER / OUTERLOOP
; This computation gets more precise the larger TIME is. A good precision measure is the difference between 
; INNERLOOP and 255.  

.set ScaledClock = CLOCK / 1000
.set DelayCallOverhead = 2					; measured in InnerLoop runs equivalent

.MACRO Delay
	ldi DelayOuterLoop, (@0 * ScaledClock / 5000 - DelayCallOverhead + 256) / 257
	ldi DelayInnerLoop, (@0  * ScaledClock / 5000 - DelayCallOverhead) / ((@0 * ScaledClock / 5000 - DelayCallOverhead + 256) / 257) - 2
	rcall ShortDelay
.ENDMACRO