; 
; MIDI Monitor Displaying 
; 
; Routines for the displaying of the buffered MIDI events
;
;
; Andreas Tnne
; atoenne@t-online.de
;
; No copyright. This is free software. Thank you for naming the author

; See the comments in MIDI Monitor Display.asm for an explanation of the 
; logical buffer organization and its relationship to the display.
;
; Web released version 1.0

; ******************** USAGE *************************
;
; .include "MIDI Monitor Buffer.asm"
; .DEF DisplayMode						0 for HEX display, 1 for SYMBOLIC display
; .DEF BufferMode						0 for CONTINUOUS, 1 for SAMPLE&HOLD
; .SET BUFFER_SIZE						The desired size of the MIDI buffer
; .DEF BufferWrite						The position in the buffer that will be written next
; .DEF DisplayTop						The first displayed byte
; .DEF DisplaySize						The number of displayed bytes
; .DEF LineSize							The number of bytes in one line
; .DEF DisplayTmp						For various tasks; unmodified by the display routines
; register Y is reserved for the buffer access
;
; rcall DisplayClear					Clears the screen and resets the displaying window
; rcall	DisplayRedisplay				Full screen redisplay. This is the fastest way even for scrolling. Surprise!

;######################################################################################
; The mapping of event status bytes to the event names and parameter displaying routines
; We distinguish channel events from broadcasting event in the same way as the receiver

NOTEOFF_L: 					.db "N.Off "
NOTEON_L:					.db "N.On  "
AFTERTOUCH_L:				.db "Aft   "
CONTROLCHANGE_L:			.db "CC    "
PROGRAMCHANGE_L:			.db "Prog  "
CHANNELPRESSURE_L:			.db "Chan  "
PITCHWHEEL_L:				.db "Pitch "
SONGPOSITION_L:				.db "S.pos "
SONGSELECT_L:				.db "S.sel "
TUNE_L:						.db "Tune  "
MIDICLOCK_L:				.db "Clock "
START_L:					.db "Start "
CONTINUE_L:					.db "Cont  "
STOP_L:						.db "Stop  "
PANIC_L:					.db "Reset "
UNKNOWN_L:					.db "Unknown "

CHANNEL_L:					.db "C "
VELOCITY_L:					.db " V  "
NOTE_L:						.db " N  "
CONTROLLER_L:				.db " Ctl  "
EQUAL_L:					.db " =  "
PROGRAM_L:					.db " P  " 

; The table of the channel event displaying routines
; The bits 6-4 of the status byte are used as an index into this table

ChannelEventRoutines:		.dw DisplaySYMBOLICNoteOff, DisplaySYMBOLICNoteOn, DisplaySYMBOLICAftertouch, DisplaySYMBOLICControlChange
							.dw DisplaySYMBOLICProgramChange, DisplaySYMBOLICChannelpressure, DisplaySYMBOLICPitchwheel

; The table of the broadcast event displaying routines
; The bits 3-0 of the status byte are used as an index into this table
; Undefined or unprintable events are shown as unknown.

BroadcastEventRoutines:		.dw DisplaySYMBOLICUnknown, DisplaySYMBOLICUnknown, DisplaySYMBOLICSongposition
							.dw DisplaySYMBOLICSongSelect, DisplaySYMBOLICUnknown, DisplaySYMBOLICUnknown, DisplaySYMBOLICTune
							.dw DisplaySYMBOLICUnknown, DisplaySYMBOLICClock, DisplaySYMBOLICUnknown, DisplaySYMBOLICStart, DisplaySYMBOLICContinue
							.dw DisplaySYMBOLICStop, DisplaySYMBOLICUnknown, DisplaySYMBOLICUnknown, DisplaySYMBOLICReset

; This table is used for the configuration page
; EventLabels is a label address (each displayed 5 chars) for each of the 16 configurable events
; TypeIndex is the mapping of these 16 event numbers to the filter table positions (starting at MAGIC)

EventLabels: .dw NOTEOFF_L, NOTEON_L, AFTERTOUCH_L, CONTROLCHANGE_L, PROGRAMCHANGE_L, CHANNELPRESSURE_L, PITCHWHEEL_L
			.dw SONGPOSITION_L, SONGSELECT_L, TUNE_L, MIDICLOCK_L, START_L, CONTINUE_L, STOP_L, PANIC_L, UNKNOWN_L

TypeIndex: .db 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x12, 0x13, 0x16, 0x18, 0x1A, 0x1B, 0x1C, 0x1F, 0x00


CursorPositions: 	.db 0x00, 0x05, 0x0A, 0x0F
					.db 0x40, 0x45, 0x4A, 0x4F
					.db 0x80, 0x85, 0x8A, 0x8F
					.db 0xC0, 0xC5, 0xCA, 0xCF


;######################################################################################


;******* DisplayClear **********
; Clears the display and resets the LCD display window

DisplayClear:
	LCDClearDisplay
	clr DisplayTop
	clr DisplaySize
	ldi DisplayMode, 1
	rjmp DisplayToggleMode				; switch from SYMBOLIC to HEX; this initializes the display and line size

;******* DisplayRedisplay *************
; Display the current display window fully including overwriting blank spaces

DisplayRedisplay:
	LCDClearDisplay						; faster than manually writing blank chars
	BufferLoadPointer DisplayTop		; start with the top byte of the current window

	cp DisplayMode, Zero				; HEX mode?
	brne DisplayRedisplaySYMBOLIC		; nope - its SYMBOLIC mode

DisplayRedisplayHEX:
	mov DisplayScratch, DisplayTmp
	rcall DisplayPos
	mov DisplayTmp, Scratch				; logical index of DisplayTop in the buffer
	; LCD420SetCursorPos 				is not necessary here since the clear display includes cursor home
	rcall DisplayHEXLine

	ldi LCDValue, 0b01000000			; Line2
	rcall LCD420SetCursorPos
	rcall DisplayHEXLine

	ldi LCDValue, 0b10000000			; Line3
	rcall LCD420SetCursorPos
	rcall DisplayHEXLine

	ldi LCDValue, 0b11000000			; Line4
	rcall LCD420SetCursorPos
	rcall DisplayHEXLine
	mov DisplayTmp, DisplayScratch
	ret

DisplayHEXLine:
	mov LCDValue, DisplayTmp
	rcall LCDWriteHex					; display the line number
	subi DisplayTmp, -6					; increment line number
	push DisplayTmp
	ldi DisplayTmp, 6
DisplayHEXLineLoop:
	LCDMoveCursorRight					; add one blank space to the left of the byte
	ld LCDValue, Y+
	cp LCDValue, Empty
	breq DisplayHEXLineNoDisplay		; dont display if its an empty byte
	rcall LCDWriteHex
	rjmp DisplayHEXLineLoopEnd
DisplayHEXLineNoDisplay:				; dont display the empty byte but space it
	LCDMoveCursorRight
	LCDMoveCursorRight
DisplayHEXLineLoopEnd:
	dec DisplayTmp
	brne DisplayHEXLineLoop
	pop DisplayTmp						; restore the current line number = byte in buffer calculated from BufferWrite
	cpi BufferPointerL, LOW(BUFFER_START+BUFFER_SIZE) ; check if the Y pointer moved out of the buffer 
	brne DisplayHEXLineEnd
	cpi BufferPointerH, HIGH(BUFFER_START+BUFFER_SIZE)
	brne DisplayHEXLineEnd
	ldi BufferPointerL, LOW(BUFFER_START)			; then wrap the Y pointer
	ldi BufferPointerH, HIGH(BUFFER_START)
DisplayHEXLineEnd:
	ret

DisplayRedisplaySYMBOLIC:
	; LCD420SetCursorPos 				is not necessary since the clear display includes cursor home
	rcall DisplaySYMBOLICEvent

	ldi LCDValue, 0b01000000			; Line2
	rcall LCD420SetCursorPos
	rcall DisplaySYMBOLICEvent

	ldi LCDValue, 0b10000000			; Line3
	rcall LCD420SetCursorPos
	rcall DisplaySYMBOLICEvent

	ldi LCDValue, 0b11000000			; Line4
	rcall LCD420SetCursorPos
	rcall DisplaySYMBOLICEvent
	ret


DisplaySYMBOLICEvent:
	ld DisplayTmp, Y+					; get status byte
	cp DisplayTmp, Empty
	breq DisplaySYMBOLICEventEnd
	mov DisplayScratch, DisplayTmp		; preserve the status byte
	cpi DisplayTmp, SYSEX				; is this a broadcast event?
	brge DisplaySYMBOLICBroadcast
DisplaySYMBOLICChannel:					; this is a channel oriented event
	andi DisplayTmp, 0x70				; mask out the event number
	swap DisplayTmp
	ldi ZL, LOW(ChannelEventRoutines*2)
	ldi ZH, HIGH(ChannelEventRoutines*2)
	rjmp DisplaySYMBOLICIndirectBranch

DisplaySYMBOLICBroadcast:
	andi DisplayTmp, 0x0F				; mask out the event number
	ldi ZL, LOW(BroadcastEventRoutines*2)
	ldi ZH, HIGH(BroadcastEventRoutines*2)

DisplaySYMBOLICIndirectBranch:
	rcall IndexZ
	ijmp								; continue with the individual event displaying address
DisplaySYMBOLICEventEnd:
	cpi BufferPointerL, LOW(BUFFER_START+BUFFER_SIZE) ; check if the Y pointer moved out of the buffer 
	brne DisplaySYMBOLICEventRet
	cpi BufferPointerH, HIGH(BUFFER_START+BUFFER_SIZE)
	brne DisplaySYMBOLICEventRet
	ldi BufferPointerL, LOW(BUFFER_START)			; then wrap the Y pointer
	ldi BufferPointerH, HIGH(BUFFER_START)
DisplaySYMBOLICEventRet:
	ret

;******************


DisplaySYMBOLICNoteOff:
	LCDTextP NOTEOFF_L, 5
	rjmp DisplaySYMBOLICNote

DisplaySYMBOLICNoteOn:
	LCDTextP NOTEON_L, 5
DisplaySYMBOLICNote:
	LCDTextP CHANNEL_L, 2				; display the channel
	mov LCDValue, DisplayScratch		; the status byte
	andi LCDValue, 0x0F					; mask out the channel number
	rcall LCDWriteHex
	LCDTextP NOTE_L, 3					; display the note number and advance Y
	ld LCDValue, Y+
	rcall LCDWriteHex
	LCDTextP VELOCITY_L, 3				; display the key velocity (we could check for vel 0 as note off in the future)
	ld LCDValue, Y+
	rcall LCDWriteHex
	rjmp DisplaySYMBOLICEventEnd		; wrap Y is necessary and return

DisplaySYMBOLICAftertouch:
	LCDTextP AFTERTOUCH_L, 4
	rjmp DisplaySYMBOLICNote			; the remainder of aftertouch is displayed like the note values

DisplaySYMBOLICControlChange:
	LCDTextP CONTROLCHANGE_L, 3
	LCDTextP CHANNEL_L, 2				; display the channel
	mov LCDValue, DisplayScratch		; the status byte
	andi LCDValue, 0x0F					; mask out the channel number
	rcall LCDWriteHex
	LCDTextP CONTROLLER_L, 5			; display the controller number
	ld LCDValue, Y+
	rcall LCDWriteHex
	LCDTextP EQUAL_L, 3					; and the controller value
	ld LCDValue, Y+
	rcall LCDWriteHex
	rjmp DisplaySYMBOLICEventEnd		; wrap Y is necessary and return

DisplaySymbolicProgramChange:
	LCDTextP PROGRAMCHANGE_L, 5
	LCDTextP CHANNEL_L, 2				; display the channel
	mov LCDValue, DisplayScratch		; the status byte
	andi LCDValue, 0x0F					; mask out the channel number
	rcall LCDWriteHex
	LCDTextP PROGRAM_L, 3				; display the program number
	ld LCDValue, Y+
	rcall LCDWriteHex
	adiw YH:YL, 1						; eat the empty third byte
	rjmp DisplaySYMBOLICEventEnd		; wrap Y is necessary and return

DisplaySYMBOLICChannelpressure:
	LCDTextP CHANNELPRESSURE_L, 4
	LCDTextP CHANNEL_L, 2				; display the channel
	mov LCDValue, DisplayScratch		; the status byte
	andi LCDValue, 0x0F					; mask out the channel number
	rcall LCDWriteHex
	LCDTextP VELOCITY_L, 3				; display the key velocity (we could check for vel 0 as note off in the future)
	ld LCDValue, Y+
	rcall LCDWriteHex
	ld LCDValue, Y+						; eat the empty third byte
	rjmp DisplaySYMBOLICEventEnd		; wrap Y is necessary and return

DisplaySYMBOLICPitchwheel:
	LCDTextP PITCHWHEEL_L, 6
	ld Scratch, Y+						; lower 7 bit
	ld Tmp, Y+							; higher 7 bit
	sbrc Tmp, 0							; move bit 8 from the higher 7 bit to the lower 7 bit
	ori Scratch, 0x80
	lsr Tmp								; move the upper 7 bits to the upper six bit position
	mov LCDValue, Tmp
	push Scratch
	rcall LCDWriteHex					; write the two hex figures 
	pop LCDValue
	rcall LCDWriteHex
	rjmp DisplaySYMBOLICEventEnd		; wrap Y is necessary and return


DisplaySYMBOLICSongposition:
	LCDTextP SONGPOSITION_L, 6
	rjmp DisplaySYMBOLICSongVal

DisplaySYMBOLICSongSelect:
	LCDTextP SONGSELECT_L, 6
DisplaySYMBOLICSongVal:
	ld LCDValue, Y+
	rcall LCDWriteHex
	ld LCDValue, Y+						; eat the empty third byte
	rjmp DisplaySYMBOLICEventEnd		; wrap Y is necessary and return

DisplaySYMBOLICTune:
	LCDTextP TUNE_L, 5
	rjmp DisplaySYMBOLICEnd

DisplaySYMBOLICClock:
	LCDTextP MIDICLOCK_L, 5
	rjmp DisplaySYMBOLICEnd

DisplaySYMBOLICStart:
	LCDTextP START_L, 5
	rjmp DisplaySYMBOLICEnd

DisplaySYMBOLICContinue:
	LCDTextP CONTINUE_L, 5
	rjmp DisplaySYMBOLICEnd

DisplaySYMBOLICStop:
	LCDTextP STOP_L, 5
	rjmp DisplaySYMBOLICEnd

DisplaySYMBOLICReset:
	LCDTextP PANIC_L, 5

DisplaySYMBOLICEnd:
	adiw YH:YL, 2						; skip the remaining bytes of the event 
	rjmp DisplaySYMBOLICEventEnd		; wrap Y is necessary and return

DisplaySYMBOLICUnknown:
	LCDTextP UNKNOWN_L, 8
	mov LCDValue, DisplayScratch
	rcall LCDWriteHex
	rjmp DisplaySYMBOLICEnd

;******* DisplayToggleMode *******
; Toggle the display mode between HEX and SYMBOLIC
; DEBUG: we need to be much more conservative with the resulting display window

DisplayToggleMode:
	ldi Scratch, 1
	eor DisplayMode, Scratch
	clr DisplaySize
	sbrc DisplayMode, 0					; skip if HEX
	subi DisplaySize, 12				; 12 byes display in SYMBOLIC
	subi DisplaySize, -24				; 24 bytes display in HEX
	mov LineSize, DisplaySize			; line size = display size / 4
	lsr LineSize
	lsr LineSize
	ret

;****** Index Z ********
; index Z (pointing to a CSEG data table) by DisplayTmp
 
IndexZ:
	add ZL, DisplayTmp
	adc ZH, Zero
	add ZL, DisplayTmp					; twice because its word access
	adc ZH, Zero
	lpm
	mov Scratch, r0
	adiw ZH:ZL,  1
	lpm
	mov ZL, Scratch
	mov ZH, r0							; Z contains indexed address
	ret

;#######################################################################################
; Configuration page
; 
; There are up to 16 configurable events

.SET MAX_CONFIGURABLE_EVENT = 15

;****** Configure ********
; Display the configuration map and then loop, reading buttons, until the configure button was pressed
; again. Use the cursor keys to scroll around and use the mode key to 

Configure:
	cli
	clr DisplayTmp
ConfigureRedisplay:
	rcall ConfigureDisplay
	rcall ConfigurePositionCursor
ConfigureLoop:
	rcall BUTTONReadWait
	sbrc BUTTONValue, 5
	rjmp ConfigureEnd

	sbrc BUTTONValue, 2
	rjmp ConfigureToggle

	sbrc BUTTONValue, 3
	rjmp ConfigureUp
	sbrc BUTTONValue, 4
	rjmp ConfigureDown
	rjmp ConfigureLoop
ConfigureEnd:
	sei
	ret

ConfigureUp:
	dec DisplayTmp
	brge RepositionCursor
	ldi DisplayTmp, 0x0F
	rjmp RepositionCursor

ConfigureDown:
	inc DisplayTmp
	andi DisplayTmp, 0x0F	
RepositionCursor:
	rcall ConfigurePositionCursor
	rjmp ConfigureLoop

; toggle the state of the current DisplayTmp indexed configurable event

ConfigureToggle:
	rcall ConfigureSetEEPromAddress
	rcall EEPromRead					; read current status
	cp EEPromValue, Zero
	brne ToggleOff
	ldi EEPromValue, MAGICBYTE
	rjmp DoToggle
ToggleOff:
	clr EEPromValue
DoToggle:
	rcall EEPromWrite
	rjmp ConfigureRedisplay


; Display the configure screen

ConfigureDisplay:
	push DisplayTmp
	LCDClearDisplay
	clr DisplayTmp						; index into the configurable events
ConfigurableDisplayLoop:				; for each configurable event
	rcall ConfigurePositionCursor
	rcall ConfigureSetEEPromAddress		; get the filter value for the current loop index
	rcall EEPromRead
	cp EEPromValue, Zero
	breq EventIsFiltered
	ldi LCDValue, '*'
	rjmp DisplayFilterStatus
EventIsFiltered:
	ldi LCDValue, ' '
DisplayFilterStatus:
	rcall LCDWriteData					; display ' ' or '*' to indicate that the next event type is filtered or not
	ldi ZL, LOW(EventLabels*2)			; get the corresponding event label
	ldi ZH, HIGH(EventLabels*2)
	rcall IndexZ
	lsl ZL
	rol ZH								; multiply by 2 for word access
	ldi LCDValue, 4
	rcall LCDTextPR
	inc DisplayTmp
	cpi DisplayTmp, MAX_CONFIGURABLE_EVENT
	brne ConfigurableDisplayLoop
	pop DisplayTmp
	ret

; load EEPromAddressL with the filter table address matching DisplayTmp

ConfigureSetEEPromAddress:
	ldi ZL, LOW(TypeIndex*2)
	ldi ZH, HIGH(TypeIndex*2)
	add ZL, DisplayTmp
	adc ZH, Zero
	lpm									; r0 = EEPROM index
	ldi EEPromAddressL, MAGIC
	add EEPromAddressL, r0
	ret

; input is DisplayTmp (0-14) the configurable event number

ConfigurePositionCursor:
	ldi ZL, LOW(CursorPositions*2)
	ldi ZH, HIGH(CursorPositions*2)
	add ZL, DisplayTmp
	adc ZH, Zero
	lpm
	mov LCDValue, r0
	rcall LCD420SetCursorPos
	ret

