; 
; MIDI Monitor Receiver
; 
; Routines for receiving and filtering MIDI events
;
;
; Andreas Tnne
; atoenne@t-online.de
;
; No copyright. This is free software. Thank you for naming the author
;
; Web released version 1.0

;*********************************************************************************
;
; Physically MIDI events are received bytewise by the UART and buffered  internally.
; The MIDI receiver picks up these bytes and interprets them as streams of MIDI
; events. MIDI events are filtered through a configurable set of rules and then 
; placed in the event buffer. Each event takes the same number of three bytes (padded by
; the Empty event value) and realtime events that are interleaved with normal events
; are moved before the normal event start. This alters the realtime event position by
; at most 650us. 


;****************** Usage *********************
;
; .DEF Byte1				The status byte of the currently received event (running status expanded)
; .DEF Byte2				The first databyte or Empty
; .DEF Byte3				The second databyte or Empty
; .DEF LastStatus			The status byte of the last received event or 0 if no event was received yet
; .DEF BytesToGo			The remaining number of bytes for the current event; negative if receiving a Sysex event
; .DEF MIDIValue			The currently processed byte; reserved register


;##########################################################################################
; A note on the MIDI event format
; All MIDI events start with a (possibly implicit) value > 127 which indicates the type of the event - the status byte
; Consecutive events with the same status byte may omit the status after the first event - called running status
; (note: we expand running status to simplify the buffer handling)
; The upper nibble of the status byte distinguishes channel oriented events from broadcast events (nibble 0b1111)
; The lower nibble of channel oriented events name the MIDI channel (0-15) of the event
; Broadcast events are divided to system events (lower nibble 0b0xxx) and realtime events (lower nibble 0b1xxx)
; Realtime events may occur even inside other events. Otherwise events must be serialized.
; All events have a fixed size except for the system exclusive event SYSEX which is of variable length and terminated
; by the end of exclusive event EOX. 

; Definitions of the MIDI event names. Note: the UNDEF events are partially used by the SMF standard and need updating one day

.SET NOTEOFF = 			0b10000000
.SET NOTEON = 			0b10010000
.SET AFTERTOUCH =		0b10100000
.SET CONTROLCHANGE =	0b10110000
.SET PROGRAMCHANGE =	0b11000000
.SET CHANNELPRESSURE = 	0b11010000
.SET PITCHWHEEL = 		0b11100000
.SET SYSEX = 			0b11110000
.SET UNDEF1 = 			0b11110001
.SET SONGPOSITION = 	0b11110010
.SET SONGSELECT = 		0b11110011
.SET UNDEF2 = 			0b11110100
.SET UNDEF3 = 			0b11110101
.SET TUNE = 			0b11110110
.SET EOX = 				0b11110111
.SET MIDICLOCK = 		0b11111000
.SET UNDEF4 = 			0b11111001
.SET START = 			0b11111010
.SET CONTINUE = 		0b11111011
.SET STOP = 			0b11111100
.SET UNDEF5 = 			0b11111101
.SET ACTIVESENSING = 	0b11111110
.SET PANIC = 			0b11111111

; Event lengths for the channel oriented events; indexed by b4-b6 of the status byte
ChannelEventLengths:
.db 3, 3, 3, 3, 2, 2, 3, 0				; note: pad byte

; Event lengths for the broadcast events; indexed by the lower nibble of the status byte
; we omitted the realtime events since their value is never read
BroadcastEventLengths:
.db -1, 1, 3, 2, 1, 1, 1, 1

;##########################################################################################
;
; Filter rules:
; The first byte is written with a magic byte. This allows to see if either the EEPROM
; has not yet been initialized or (dread!) the infamous EEPROM corruption occurred. 
; The seven channel oriented events and the sixteen broadcast events are switchable through
; the table FILTERED_EVENTS, which is indexed by event byte. A value 0 means the event
; is rejected and a value > 0 means the event is allowed. Note that SYSEX and the Empty event
; (typically ACTIVE_SENSING) are always filtered irregardless of these settings

.SET MAGICBYTE = 0x0F
.eseg
MAGIC: .byte 1
FILTERED_EVENTS: .byte 32
.cseg

;##########################################################################################

;****** ReceiveMIDIValue **********
; Process the byte in MIDIValue
; - Filter Sysex and other unwanted events
; - Collect the bytes of one event in the registers
; - Store the event as 1-3 bytes in the buffer when the event is received completely
; - Trigger redisplay when the event is received completely
; Note: orphaned bytes without status byte are dropped
; Note: incomplete events are dropped

ReceiveMIDIValue:
	cp MIDIValue, Empty						; filter out the Empty value (typically ACTIVESENSING)
	breq ReceiveMIDIValueEnd
	sbrs MIDIValue, 7						; is this a status byte? If not ..
	rjmp ReceiveDataByte					; process a data byte (possibly with running status)
	cpi MIDIValue, SYSEX					; is this a broadcast event?
	brge ReceiveBroadcast
ReceiveStatusByte:							; MIDIValue is a status byte of a channel oriented event
	mov LastStatus, MIDIValue
	mov Byte1, MIDIValue
	mov Byte2, Empty
	mov Byte3, Empty
	andi MIDIValue, 0x70					; filter out status byte bit and channel information
	swap MIDIValue							; and move the event number to the lower nibble
	ldi ZL, LOW(ChannelEventLengths*2)		; get the number of expected remaining bytes
	ldi ZH, HIGH(ChannelEventLengths*2)
	add ZL, MIDIValue
	adc ZH, Zero
	lpm
	dec r0
	mov BytesToGo, r0
	inc MIDIValue							; = event type between 1 and 8
	mov EventType, MIDIValue				; remember this index of the event filtering
ReceiveMIDIValueEnd:
	ret


;***** ReceiveBroadcast ******
; MIDIValue is the status byte of a broadcast event

ReceiveBroadcast:
	cpi MIDIValue, SYSEX					; special treatment of system exclusive event
	breq ReceiveSYSEX
	cpi MIDIValue, EOX						; EOX is part of a SYSEX event and hence dropped
	breq ReceiveMidiValueEnd				
	cpi MIDIValue, MIDICLOCK				; is this a realtime event?
	brge FilterAndBufferRealtime			; then pass it into the buffer directly. Note: this alters the order of events slightly
	clr LastStatus							; invalidate running status
	mov Byte1, MIDIValue
	mov Byte2, Empty
	mov Byte3, Empty
	andi MIDIValue, 0x0F					; the lower nibble determines the kind of event
	ldi ZL, LOW(BroadcastEventLengths*2)	; get the number of expected remaining bytes
	ldi ZH, HIGH(BroadcastEventLengths*2)
	add ZL, MIDIValue
	adc ZH, Zero
	lpm
	dec r0
	mov BytesToGo, r0
	ori MIDIValue, 0x10						; = event type between 0x10 and 0x1F
	mov EventType, MIDIValue				; remember this index of the event filtering
	ret


;***** ReceiveDataByte *******
; MIDIValue is a data byte

ReceiveDataByte:
	cp BytesToGo, Zero
	brmi ReceiveMIDIValueEnd				; drop the byte if the bytes count is negative -> we are in a SYSEX event
	brne ReceiveDataByte1					; byte count = 0 -> last event is finished. Then we miss the status byte

	cp LastStatus, Zero						; do we have a running status or is this an orphaned byte?
	breq ReceiveMIDIValueEnd				; orphaned bytes are ignored
	mov Tmp, MIDIValue						; fake the receiving of the running status byte
	mov MIDIValue, LastStatus
	rcall ReceiveStatusByte
	mov MIDIValue, Tmp
ReceiveDataByte1:							; store the byte in either Byte2 or Byte3
	cp Byte2, Empty							; is this already the third byte of an event?
	brne ReceiveDataByte2
	mov Byte2, MIDIValue
	rjmp ReceiveDataByteFinish
ReceiveDataByte2:
	mov Byte3, MIDIValue
ReceiveDataByteFinish:						; check if the event is complete with this byte
	dec BytesToGo
	breq FilterAndBuffer					; then move the completed event to the buffer
	ret



;***** ReceiveSYSEX *******
; Dispose all following bytes except realtime events until any not-realtime event is received
; EOX is preferred but for robustness reasons not strictly required.
; SYSEX mode is indicated by a negative byte count. All non-status bytes are dropped as long as
; the byte count is negative. The byte count is reset only when a status byte other than EOX is received.
; This shows why EOX is not strictly required in the MIDI protocol: all databytes that are not status bytes
; are dropped after a SYSEX was finished with EOX because running status is not active and also because 
; the byte count is still negative. So we can also finish SYSEX with a true status byte and have the same result.

ReceiveSYSEX:
	clr LastStatus							; invalidate running status
	ldi Scratch, -1
	mov BytesToGo, Scratch
	ret


;***************************** FILTERING *******************************************

;****** FilterAndBufferRealtime **********
; Filter the realtime event byte in MIDIValue and put it into the buffer immediately
; If a realtime event occurs while a multi-byte event is received then this realtime
; event will be inserted into the buffer BEFORE the current event. This is done to
; simplify the logical buffer structure and displaying.

FilterAndBufferRealtime:
	rcall FilterEvent
	breq FilterAndBufferEnd

	rcall BufferWrite					; write the realtime event
	mov MIDIValue, Empty				; and pad with two empty bytes
	rcall BufferWrite
	mov MIDIValue, Empty
	rcall BufferWrite
	rjmp UpdateBufferDisplay

;******* FilterAndBuffer ********
; Filter the midi event in Byte1-3 and put it into the buffer.

FilterAndBuffer:
	rcall FilterEvent
	breq FilterAndBufferEnd

	mov MIDIValue, Byte1
	rcall BufferWrite
	mov MIDIValue, Byte2
	rcall BufferWrite
	mov MIDIValue, Byte3
	rcall BufferWrite

UpdateBufferDisplay:
	sbrc UartBufferStatus, UART_BUFFER_EMPTY_BIT	; do not display if there are still bytes to be read => overrun protection
	rcall DisplayRedisplay				; simplest solutions = write over the whole display
FilterAndBufferEnd:
	ret

;****** EEPromInit *********
; Write the initial EEPROM event filter table

EEPromInit:
	ldi EEPromAddressL, MAGIC
	ldi EEPromValue, MAGICBYTE
EEPromInitLoop:
	rcall EEPromWrite
	inc EEPromAddressL
	cpi EEPromAddressL, 0x20
	brne EEPromInitLoop
	ret

;**** FilterEvent ******
; Reject the current event (determined by the Byte1-3 and EventType) if its filter table value is zero
; return Z flag set if rejected

FilterEvent:
	ldi EEPromAddressL, MAGIC			; event type 0-7 (for channel events) are +1, @0 = MAGICBYTE
	add EEPromAddressL, EventType
	rcall EEPromRead
	cp EEPromValue, Zero
	ret									; ret Z set if rejected